/* k3 - a simple prettifier for /proc/meminfo.
 * Public domain by Russell Marks 1993-2022.
 */

/* As of 1.6, requires C99 */

#include <stdio.h>
#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


#define K3_VERSION	"1.6"


/* used via multiply/divide anyway, so using large int */
intmax_t buf_free=1;


#define BAR_WIDTH	40

#define USEDCHAR '#'
#define DBUFCHAR '%'
#define FREECHAR ':'


/* Yes, it does still work in KB internally, and yes this is a bit silly.
 * I nearly added a "-k" to show sizes in KB still, but KB hasn't
 * really been the right choice since maybe the late 90s, so, y'know.
 */
void showbar(char *title,intmax_t declen,
             intmax_t availk,intmax_t buffk,intmax_t totalk)
{
intmax_t limit1,limit2,count,pcnt;

pcnt=((availk+buf_free*buffk)*100)/totalk;

printf("%s ",title);
printf("%3jd%% free (%*jdM/%*jdM) [",
       pcnt,
       (int)declen,(availk+buf_free*buffk)/1024,
       (int)declen,totalk/1024);

limit1=BAR_WIDTH-((availk*BAR_WIDTH)/totalk);
limit2=BAR_WIDTH-(((availk+buffk)*BAR_WIDTH)/totalk);

for(count=0;count<BAR_WIDTH;count++)
  putchar((count<limit2)?USEDCHAR:((count<limit1)?DBUFCHAR:FREECHAR));

puts("]");
}


/* works for 2.0?, 2.2, 2.4.
 *
 * (I don't know if the C99 requirement makes it entirely pointless this
 * still being here now or not - it might do.)
 */
void readmeminfo_old(intmax_t *pmf,intmax_t *pmt,intmax_t *pmb,
                     intmax_t *pms,intmax_t *psf,intmax_t *pst)
{
FILE *in;
char buf[128],junk[128];
intmax_t waste,cached;

if((in=fopen("/proc/meminfo","r"))==NULL)
  fprintf(stderr,"k3: couldn't read /proc/meminfo.\n"),exit(1);

fgets(buf,sizeof(buf),in);
fgets(buf,sizeof(buf),in);
sscanf(buf,"%s%jd%jd%jd%jd%jd%jd",junk,pmt,&waste,pmf,pms,pmb,&cached);
/* we bundle together buffers/cached like `free' does with its "+/-"... */
*pmb+=cached;
fgets(buf,sizeof(buf),in);
sscanf(buf,"%s%jd%jd%jd",junk,pst,&waste,psf);
fclose(in);

/* actually want it in k, not bytes */
*pmf>>=10;
*pmt>>=10;
*pmb>>=10;
*pms>>=10;
*psf>>=10;
*pst>>=10;
}


/* works for 2.2?, 2.4+; reads the "kB" fields.
 * The advantage is that this copes with >=2GB (or is it 4GB?) RAM or
 * swap values correctly on 32-bit systems.
 * Returns 0 if it failed and we should try the old scheme.
 */
int readmeminfo_new(intmax_t *pmf,intmax_t *pmt,intmax_t *pmb,
                    intmax_t *pms,intmax_t *psf,intmax_t *pst)
{
FILE *in;
char buf[128],*ptr;
intmax_t cached=0,val;

if((in=fopen("/proc/meminfo","r"))==NULL)
  fprintf(stderr,"k3: couldn't read /proc/meminfo.\n"),exit(1);

*pmt=0;

while(fgets(buf,sizeof(buf),in)!=NULL)
  {
  ptr=strchr(buf,':');
  if(!ptr) continue;
  *ptr=0;
  val=atoll(ptr+1);
  
  if(strcmp(buf,"MemTotal")==0)
    *pmt=val;
  else if(strcmp(buf,"MemFree")==0)
    *pmf=val;
  else if(strcmp(buf,"MemShared")==0)
    *pms=val;
  else if(strcmp(buf,"Buffers")==0)
    *pmb=val;
  else if(strcmp(buf,"Cached")==0)
    cached=val;
  else if(strcmp(buf,"SwapTotal")==0)
    *pst=val;
  else if(strcmp(buf,"SwapFree")==0)
    *psf=val;
  }

fclose(in);

if(*pmt==0) return(0);		/* failed */

/* bundle together buffers/cached */
*pmb+=cached;

return(1);
}


void usage(void)
{
printf("k3 " K3_VERSION " - public domain by Russell Marks.\n\n");
puts("usage: k3 [-r]\n"
"\n"
"	-r	show real amount free (in numeric totals), rather than \n"
"		regarding buffers/cached as more-or-less `free'.");
}


/* get length of number in decimal digits.
 *
 * I know log10() exists, but I wanted to avoid FP and any possible
 * accuracy issues that might arise from that. Speed is not really
 * an issue here, so integer brute-force works for me. :-)
 *
 * (Return value being intmax_t is excessive, but potentially avoids
 * confusion.)
 */
intmax_t maxdeclen(intmax_t num)
{
intmax_t declog=0;

while(num>0) declog++,num/=10;

return(declog);
}


int main(int argc,char *argv[])
{
intmax_t mf,mt,mb,ms,sf,st,len_mt,len_st,len_max;

if(argc==2 && strcmp(argv[1],"-r")==0)
  buf_free=0,argc--,argv++;

if(argc!=1) usage(),exit(1);

if(!readmeminfo_new(&mf,&mt,&mb,&ms,&sf,&st))
  readmeminfo_old(&mf,&mt,&mb,&ms,&sf,&st);

/* minus 3 to allow for MB output */
len_mt=maxdeclen(mt)-3;
len_st=maxdeclen(st)-3;
len_max=(len_mt>len_st?len_mt:len_st);

/* impossible in practice? - but just in case */
if(len_max<1) len_max=1;

showbar("Phys",len_max,mf,mb,mt);
if(st>0)
  showbar("Swap",len_max,sf,0,st);

exit(0);
}
